/* Text clustering method, SKWIC algorithm implementation
 * Model class containing methods necessary for pre-processing
 * Authors: Roman Milos, Vladimir Bartik, May 2011
 */
package dip.model;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import lib.PorterStemmer;

/**
 * Model class containing methods necessary for pre-processing
 * @author Roman
 */
public class Predzpracovani {

    // singleton instance creation
    private static Predzpracovani instance = new Predzpracovani();

    //getting a singleton instance
    public static Predzpracovani getInstance() {
        return Predzpracovani.instance;
    }

    /**
     * Constructor - get all necessary data and initialize a stemmer
     */
    private Predzpracovani() {
        data = Data.getInstance();
        s = new PorterStemmer();
    }
    //variables necessary to run the application
    private Data data;
    private PorterStemmer s;
    private String aktualniSlovo;
    private String text;
    private int citacParsovani = 0;
    private int citacVahovani = 0;
    private String postupText;
    private boolean zmena = true;
    public boolean storno = false;
    private long cas = 0;

    /**
     * Returns the count of documents
     * @return count of documents
     */
    public int getCelkem() {

        return data.getDokumenty().size();

    }

    /**
     * Returns the running time of pre-processing
     * @return pre-processing running time
     */
    public long getCas() {
        return cas;
    }

    /**
     * Counts and returns general information about the progress
     * @return number characterizing the process
     */
    public int getPostup() {

        return (750 * citacParsovani + 250 * citacVahovani) / (getCelkem() + 1);

    }

    /**
     * Returns the actual progress information of pre-processing
     * @return string containing the actual state
     */
    public String getPostupText() {
        return postupText;
    }

    /**
     * Returns, if something was changed in the class
     * @return boolean value
     */
    public boolean isZmena() {
        return zmena;
    }
    
    /**
     * Sets the "change" boolean value
     * @param zmena boolean value
     */
    public void setZmena(boolean zmena) {
        this.zmena = zmena;
    }

    /**
     * Returns the dataset used for pre-processing
     * @return the whole dataset
     */
    public Data getData() {
        return data;
    }

    /**
     * Sets the dataset used for pre-processing
     * @param data the input dataset
     */
    public void setData(Data data) {
        this.data = data;
    }

    public void testSouboru() {
    }


    /**
     * Main method used to execute pre-processing, sets the execution time, 
     * progress and all the information and executes all pre-processing methods. 
     * @throws Exception
     */
    public void beh() throws Exception {

        storno = false;
        if (zmena) {
            cas = -System.currentTimeMillis();
            data.clean();
            //  System.gc();
            citacParsovani = 0;
            citacVahovani = 0;

            postupText = "Reading stopList...";

            nahrajStopList();

            postupText = "Reading data collection...";
            if (data.getDataFile().isFile()) {
                ziskejKolekciDatZeSouboru();
            } else if (data.getDataFile().isDirectory()) {
                ziskejKolekciDatZeSlozky();
            } else {
                throw new Exception("Unknown type of input data");

            }

            data.getSlovaDokumenty().clear();


            for (Dokument d : data.getDokumenty()) {
                dokumentParse(d);

                citacParsovani++;
                if (storno) {
                    return;
                }
                postupText = "Parsing: " + citacParsovani + " from " + data.getDokumenty().size();

            }
            postupText = "Selection of most frequent terms";
            odstanJednoDokumentoveTermy();
            vyberNejcastejsiTermy();
            odstranPrazdneDokumenty();

            if (data.getDokumenty().size() < 1) {
                throw new Exception("Document folder does not contain any files with data");

            }
            ziskejFrekvence();
            for (Dokument d : data.getDokumenty()) {
                //       d.setSoubor(null);

                dokumentVahuj(d);

                citacVahovani++;

                if (storno) {
                    return;
                }
                postupText = "Weighting: " + citacVahovani + " from " + data.getDokumenty().size();


            }
        }


        opravID();
        setZmena(false);

        System.gc();
        cas += System.currentTimeMillis();
    }

    /**
     * Removes all empty documents in the dataset.
     */
    private void odstranPrazdneDokumenty() {
        List<Dokument> kOdstraneni = new ArrayList<Dokument>();
        for (Dokument d : data.getDokumenty()) {

            if (d.getSlova().size() < 2) {
                kOdstraneni.add(d);
            }
        }
        for (Dokument d : kOdstraneni) {
            data.getDokumenty().remove(d);
        }
    }

    /**
     * Selects the most frequent terms. If the count of terms will be higher
     * than required, it is OK, otherwise it removes some unnecessary terms. 
     */
    public void vyberNejcastejsiTermy() {
        if (data.getSlovaDokumenty().size() < data.getMaximalneTermu()) {
            return;
        }
        List<SlovoSuma> slova = new ArrayList<SlovoSuma>(data.getSlovaDokumenty().values());
        Collections.sort(slova, new PorovnejSlovoSumaDleIDF());
        for (int i = 0; i < slova.size() - data.getMaximalneTermu(); i++) {
            data.getSlovaDokumenty().remove(slova.get(i).getJmeno());

        }
        for (Dokument d : data.getDokumenty()) {
            Slovo slovo;
            for (Iterator<Slovo> iter = d.getSlova().values().iterator(); iter.hasNext();) {
                slovo = iter.next();
                if (!data.getSlovaDokumenty().containsKey(slovo.getJmeno())) {
                    iter.remove();

                }
            }
        }
    }

    /**
     * Class used to compare two weight values of two words
     */
    private class PorovnejSlovoSumaDleIDF implements java.util.Comparator<SlovoSuma> {

        @Override
        public int compare(SlovoSuma o1, SlovoSuma o2) {
            Integer sl1 = ((SlovoSuma) o1).getPocetDokumentu();
            Integer sl2 = ((SlovoSuma) o2).getPocetDokumentu();
            return sl1.compareTo(sl2);
        }
    }

    /**
     * Calls a suitable weighting method for a given document.
     * @param dokument an input document
     */
    public void dokumentVahuj(Dokument dokument) {

        switch (data.getVahovani()) {
            case IDF:
                vahujIDF(dokument);
                break;
            case TF:
                vahujTF(dokument);
                break;
            default:
                vahujTFIDF(dokument);

        }
    }

    /**
     * Counts the TF weighs for a given document, including normalization. 
     * @param dokument an input document
     */
    public void vahujTF(Dokument dokument) {
        double pocet;

        double suma = 0.0;
        for (Slovo sl : dokument.getSlova().values()) {
            pocet = (double) sl.getPocet() / data.getSlovaDokumenty().get(sl.getJmeno()).getMaximalniVyskyty();
            sl.setVaha(pocet);
            suma += pocet;
        }
        for (Slovo sl : dokument.getSlova().values()) {
            pocet = sl.getPocet();
            sl.setVaha(pocet / suma);
        }

    }

    /**
     * Counts the IDF weighs for a given document, including normalization. 
     * @param dokument an input document
     */
    public void vahujIDF(Dokument dokument) {
        double idf;
        double suma = 0.0;
        SlovoSuma slovo;
        for (Slovo sl : dokument.getSlova().values()) {
            slovo = data.getSlovaDokumenty().get(sl.getJmeno());

            idf = (((double) data.getDokumenty().size()) / slovo.getPocetDokumentu());
            idf = Math.log(idf);

            suma += idf;
            sl.setVaha(idf);
        }
        for (Slovo sl : dokument.getSlova().values()) {
            idf = sl.getVaha();
            sl.setVaha(idf / suma);
        }
    }

    /**
     * Counts the TF-IDF weighs for a given document, including normalization. 
     * @param dokument an input document
     */
    public void vahujTFIDF(Dokument dokument) {
        double tfidf;
        double suma = 0.0;
        double tf;
        double idf;
        SlovoSuma slovo;
        for (Slovo sl : dokument.getSlova().values()) {
            slovo = data.getSlovaDokumenty().get(sl.getJmeno());
            tf = (double) sl.getPocet() / data.getSlovaDokumenty().get(sl.getJmeno()).getMaximalniVyskyty();
            idf = Math.log10(((double) data.getDokumenty().size()) / slovo.getPocetDokumentu());
            tfidf = tf * idf;
            suma += tfidf;
            sl.setVaha(tfidf);
        }
        for (Slovo sl : dokument.getSlova().values()) {
            tfidf = sl.getVaha();
            sl.setVaha(tfidf / suma);

        }
    }

    /**
     * Removes all terms, which are contained in only one document.
     */
    public void odstanJednoDokumentoveTermy() {
        if (data.isMazatTermyJednohoDokumentu()) {
            Collection<SlovoSuma> kOdstraneni = new ArrayList<SlovoSuma>();
            for (SlovoSuma slovo : data.getSlovaDokumenty().values()) {
                if (slovo.getPocetDokumentu() == 1) {

                    kOdstraneni.add(slovo);
                    slovo.getDokument().getSlova().remove(slovo.getJmeno());
                }
            }
            for (SlovoSuma slovo : kOdstraneni) {
                data.getSlovaDokumenty().remove(slovo.getJmeno());
            }
        }
    }

    /**
     * If stemming has to be used, executes Porter stemmer and returns 
     * its results.
     * @param slovo a word to be stemmed
     * @return resulting word
     */
    public String stemm(String slovo) {

        if (data.isAplikovatStemming()) {
            return s.stem(slovo);
        } else {
            return slovo;
        }

    }

    /**
     * If stop words removal is disabled, returns false, otherwise it is 
     * checked, if it is a stop word.
     * @param slovo an input word
     * @return a boolean value, true if an input word is a stop word
     */
    public boolean jeStopSlovo(String slovo) {
        if (data.isAplikovatStopList()) {
            return data.getStopSlova().containsKey(slovo);
        } else {
            return false;
        }
    }

    /**
     * If number removal is disabled, returns false, otherwise it is 
     * checked, if it is a number.
     * @param slovo an input word
     * @return a boolean value, true if an input word is a number
     */
    public boolean jeCislo(String slovo) {
        if (data.isMazatCisla()) {
            return slovo.matches("^[0-9].*$");
        } else {
            return false;
        }
    }

    /**
     * Gets a stoplist and stores all stop words into a hash map.
     * @throws Exception
     */
    private void nahrajStopList() throws Exception {
        if (!data.isAplikovatStopList()) {
            return;
        }
        File soubor = data.getStopListFile();
        data.testStopListFile(soubor);
        try {

            text = IO.souborNaString(soubor);
            String[] slova = textToSlova(text);
            for (int i = 0; i < slova.length; i++) {
                data.getStopSlova().put(slova[i], true);

            }

        } catch (java.io.IOException e) {
            throw new Exception("Unable to read stop words file");
        } catch (Exception e) {
            throw e;
        }
        if (data.getStopSlova().size() < 1) {

            throw new Exception("Empty stop words file");
        }
    }

    /**
     * Divides the input text into words, according to the white characters.
     * @param text input string with text
     * @return array of words
     */
    private String[] textToSlova(String text) {
        return text.split("(\\s)+");
    }

    /**
     * Replace all characters except alphanumeric characters and dots with 
     * a space character. 
     * @param textStary input text string
     * @return resulting text string
     */
    private String vycistiText(String textStary) {
        return textStary.replaceAll("&.*;|[^a-zA-Z0-9'\\.\n]", " ");
    }

   /**
     * Replace all dots with a space character. 
     * @param textStary input text string
     * @return resulting text string
     */
    private String vycistiSlovo(String slovoKCisteni) {
        return slovoKCisteni.replaceAll("\\.|'", "");
    }

    /**
     * Gets the maximum frequency of a term in the whole text collection. 
     */
    public void ziskejFrekvence() {
        SlovoSuma slovoSuma;
        for (Dokument d : data.getDokumenty()) {

            for (Slovo slovo : d.getSlova().values()) {
                slovoSuma = data.getSlovaDokumenty().get(slovo.getJmeno());
                if (slovoSuma.getMaximalniVyskyty() < slovo.getPocet()) {
                    slovoSuma.setMaximalniVyskyty(slovo.getPocet());
                }
            }
        }

    }

    /**
     * Sets all identifiers of terms so as it can be used as an array index. 
     */
    public void opravID() {
        int id = 0;

        for (SlovoSuma slovo : data.getSlovaDokumenty().values()) {
            slovo.setId(id);
            id++;
        }
        for (Dokument d : data.getDokumenty()) {
            id = 0;
            for (Slovo slovo : d.getSlova().values()) {
                slovo.setId(id);
                id++;
            }
        }


    }

    /**
     * Gets the document information from a file. Assigns the category, title,
     * name and text for it.
     * @throws IOException
     */
    public void ziskejKolekciDatZeSouboru() throws IOException {
        Dokument d;

        String obsahSouboru = IO.souborNaString(data.getDataFile());
        String[] soubory = obsahSouboru.split("\n");
        String[] textSouboru;
        for (int i = 0; i <
                soubory.length; i++) {
            textSouboru = soubory[i].split("\\t", 2);
            if (textSouboru.length == 2) {
                d = new Dokument();
                d.setJmeno("" + i);
                // d.setTitle(i + " " + textSouboru[0]);
                d.setText(textSouboru[1]);
                d.setKategorie(textSouboru[0]);
                data.getKategorie().put(textSouboru[0], 0);
                data.getDokumenty().add(d);
                // System.out.println(soubory[i]);System.exit(i);

            } else if (textSouboru.length == 1) {
                d = new Dokument();
                d.setJmeno("" + i);
                d.setTitle("" + i);
                d.setText(soubory[i]);
                d.setKategorie("");
                data.getDokumenty().add(d);

            }

        }

        if (data.getPocetTrid() < 2) {
            data.setPocetTrid(data.getKategorie().size());


        }

    }

    /**
     * Takes a folder of document files and stores them as documents (including 
     * title, category and text).
     * @throws Exception
     */
    public void ziskejKolekciDatZeSlozky() throws Exception {

        String aktualniDokument;
        File soubor;

        File prohledavany = data.getDataFile();
        List<File> soubory = new ArrayList<File>();
        soubory.add(prohledavany);
        Dokument dokument;

        data.getDokumenty().clear();
        for (int i = 0; i <
                soubory.size(); i++) {

            prohledavany = soubory.get(i);

            String[] jmenaSouboru = prohledavany.list();
            ArrayList keZpracovani = new ArrayList<String>(Arrays.asList(jmenaSouboru));
            Iterator it = keZpracovani.iterator();
            while (it.hasNext()) {

                aktualniDokument = it.next().toString();
                try {
                    dokument = new Dokument();

                    soubor =
                            new File(prohledavany.getAbsolutePath() + "\\" + aktualniDokument);

                    if (soubor.isDirectory() && soubor.canRead()) {
                        soubory.add(soubor);
                        continue;

                    } else if (!(soubor.exists() && soubor.isFile() && soubor.canRead())) {
                        //System.out.println(soubor.getAbsolutePath());
                        continue;
                    }

                    dokument = new Dokument();
                    dokument.setJmeno(aktualniDokument);
                    dokument.setSoubor(soubor);
                    text =
                            IO.souborNaString(dokument.getSoubor());
                    dokument.setText(text);
                    try {
                        dokument.setTitle(text.split("\n", 2)[0]);
                    } catch (Exception e) {
                        dokument.setTitle("");
                    }

                    data.getDokumenty().add(dokument);

                } catch (Exception e) {
                    // throw e;
                }
            }
        }

    }

    /**
     * Parses a document into a set of terms and performs all necessary 
     * pre-processing methods.
     * @param dokument an input document
     * @throws IOException
     */
    public void dokumentParse(Dokument dokument) throws IOException {

        SlovoSuma slovoSum;
        dokument.getSlova().clear();


        text =
                vycistiText(dokument.getText().toLowerCase());

        String[] slova = textToSlova(text);
        text =
                null;
        Slovo slovo;

        for (int i = 0; i <
                slova.length; i++) {
            aktualniSlovo = slova[i];
            aktualniSlovo =
            stemm(vycistiSlovo(aktualniSlovo));
            if (jeStopSlovo(aktualniSlovo)) {
                continue;
            }

            if (jeCislo(aktualniSlovo)) {
                continue;
            }

            if (aktualniSlovo.length() < 2) {
                continue;
            }

            slovo = dokument.getSlova().get(aktualniSlovo);
            if (slovo != null) {
                slovo.inc();
                data.getSlovaDokumenty().get(aktualniSlovo).inc();

            } else {
                slovo = new Slovo(aktualniSlovo);
                slovo.setId(dokument.getSlova().size());
                dokument.getSlova().put(aktualniSlovo, slovo);
                slovoSum =
                        data.getSlovaDokumenty().get(aktualniSlovo);
                if (slovoSum != null) {
                    slovoSum.incDokument(dokument);

                } else {
                    slovoSum = new SlovoSuma(aktualniSlovo);
                    slovoSum.setId(data.getSlovaDokumenty().size());
                    slovoSum.setDokument(dokument);
                    data.getSlovaDokumenty().put(aktualniSlovo, slovoSum);


                }
            }
        }


    }
}
